@@ -1,6 +1,5 @@
 #! /bin/sh
 
-# <pre>
 # This file is in the public domain, so clarified as of
 # 2009-05-17 by Arthur David Olson.
 
@@ -14,6 +13,8 @@ echo ".am TH
 ..
 .rm }H
 .rm }F" | nroff -man - ${1+"$@"} | perl -ne '
+	binmode STDIN, '\'':encoding(utf8)'\'';
+	binmode STDOUT, '\'':encoding(utf8)'\'';
 	chomp;
 	s/.\010//g;
 	s/\s*$//;
